from .nonlinear import NonlinearSystem
import numpy as np

class LinearSystem(NonlinearSystem):
    """
    A linear discrete-time system of the form

    .. math::

        x_{t+1} = A x_t + B u_t

        y_t = C x_t + D u_t

    where

        - :math:`x_t \in \mathbb{R}^n` is a system state,
        - :math:`u_t \in \mathbb{R}^m` is a control input,
        - :math:`y_t \in \mathbb{R}^p` is a system output.

    :param A: A ``(n,n)`` numpy array representing the state transition matrix
    :param B: A ``(n,m)`` numpy array representing the control input matrix
    :param C: A ``(p,n)`` numpy array representing the state output matrix
    :param D: A ``(p,m)`` numpy array representing the control output matrix
    """
    def __init__(self, A, B, C, D):
        self.n = A.shape[1]
        self.m = B.shape[1]
        self.p = C.shape[0]

        # Sanity checks on matrix sizes
        assert A.shape == (self.n, self.n), "A must be an (n,n) matrix"
        assert B.shape == (self.n, self.m), "B must be an (n,m) matrix"
        assert C.shape == (self.p, self.n), "C must be an (p,n) matrix"
        assert D.shape == (self.p, self.m), "D must be an (p,m) matrix"

        # Store dynamics parameters
        self.A = A
        self.B = B
        self.C = C
        self.D = D

        # Dynamics functions
        self.dynamics_fcn = lambda x, u: A@x + B@u
        self.output_fcn = lambda x, u: C@x + D@u

class DoubleIntegrator(LinearSystem):
    """
    A linear system describing a double integrator in :math:`d` dimensions
    with full state and control output:

    .. math::

        A = \\begin{bmatrix} I_{d \\times d}  & I_{d \\times d} \\\ 0_{d \\times d} & I_{d \\times d}  \\end{bmatrix}
        \quad
        B = \\begin{bmatrix} 0_{d \\times d} \\\ I_{d \\times d}  \\end{bmatrix}

    .. math::
        C = \\begin{bmatrix} I_{2d \\times 2d} \\\ 0_{d \\times 2d} \\end{bmatrix}
        \quad
        D = \\begin{bmatrix} 0_{2d \\times d} \\\ I_{d \\times d} \\end{bmatrix}

    :param d: Integer describing the dimensionality of the system
    """
    def __init__(self, d):
        I = np.eye(d)
        z = np.zeros((d,d))

        A = np.block([[I,I],
                      [z,I]])
        B = np.block([[z],
                      [I]])
        C = np.block([[I,z],
                      [z,I],
                      [z,z]])
        D = np.block([[z],
                      [z],
                      [I]])

        LinearSystem.__init__(self, A, B, C, D)



from .base import BenchmarkScenario
from .common import (inside_rectangle_formula,
                     outside_rectangle_formula,
                     make_rectangle_patch)
from ..systems import DoubleIntegrator

class EitherOr(BenchmarkScenario):
    r"""
    A 2D mobile robot with double integrator dynamics must
    avoid an obstacle (:math:`\mathcal{O}`) before reaching a goal
    (:math:`\mathcal{G}`). Along the way, the robot must reach one
    of two intermediate targets (:math:`\mathcal{T}_i`) and stay
    there for several timesteps:

    .. math::

        \varphi = 
            F_{[0,T-\tau]} 
                \left( G_{[0,\tau]} \mathcal{T}_1 \lor G_{[0,\tau]} \mathcal{T}_2 \right)
            \land F_{[0,T]} \mathcal{G} 
            \land G_{[0,T]} \lnot \mathcal{O}

    :param goal:        Tuple containing bounds of the rectangular goal region
    :param target_one:  Tuple containing bounds of the rectangular first target
    :param target_two:  Tuple containing bounds of the rectangular second target
    :param obstacle:    Tuple containing bounds of the rectangular obstacle
    :param T:           Total number of time steps
    :param T_dwell:     Dwell time :math:`\tau` (integer number of timesteps)
    """
    def __init__(self, goal, target_one, target_two, obstacle, T, T_dwell):
        self.goal = goal
        self.target_one = target_one
        self.target_two = target_two
        self.obstacle = obstacle
        self.T = T
        self.T_dwell = T_dwell

    def GetSpecification(self):
        # Goal Reaching
        at_goal = inside_rectangle_formula(self.goal, 0, 1, 6)

        # Target reaching
        at_target_one = inside_rectangle_formula(self.target_one, 0, 1, 6).always(0, self.T_dwell)
        at_target_two = inside_rectangle_formula(self.target_two, 0, 1, 6).always(0, self.T_dwell)
        at_either_target = at_target_one | at_target_two

        # Obstacle Avoidance
        not_at_obstacle = outside_rectangle_formula(self.obstacle, 0, 1, 6)

        specification = at_either_target.eventually(0, self.T-self.T_dwell) & \
                        not_at_obstacle.always(0, self.T) & \
                        at_goal.eventually(0, self.T)

        return specification

    def GetSystem(self):
        return DoubleIntegrator(2)

    def add_to_plot(self, ax):
        # Make and add rectangular patches
        ax.add_patch(make_rectangle_patch(*self.obstacle, color='k', alpha=0.5))
        ax.add_patch(make_rectangle_patch(*self.target_one, color='blue', alpha=0.5))
        ax.add_patch(make_rectangle_patch(*self.target_two, color='blue', alpha=0.5))
        ax.add_patch(make_rectangle_patch(*self.goal, color='green', alpha=0.5))

        # set the field of view
        ax.set_xlim((0,10))
        ax.set_ylim((0,10))
        ax.set_aspect('equal')



from .drake_micp import DrakeMICPSolver
from ...STL import LinearPredicate
import numpy as np
from pydrake.all import (AddLogarithmicSos1Constraint,
                         eq, le, ge)

class DrakeSos1Solver(DrakeMICPSolver):
    """
    Given an :class:`.STLFormula` :math:`\\varphi` and a :class:`.LinearSystem`,
    solve the optimization problem

    .. math::

        \min & -\\rho^{\\varphi}(y_0,y_1,\dots,y_T) + \sum_{t=0}^T x_t^TQx_t + u_t^TRu_t

        \\text{s.t. } & x_0 \\text{ fixed}

        & x_{t+1} = A x_t + B u_t

        & y_{t} = C x_t + D u_t

        & \\rho^{\\varphi}(y_0,y_1,\dots,y_T) \geq 0

    using mixed-integer convex programming. This method uses fewer binary variables
    by encoding disjunction with a Specially Ordered Set of Type 1 (SOS1) constraint.

    .. note::

        This class implements the encoding described in

        Kurtz V, et al.
        *Mixed-Integer Programming for Signal Temporal Logic with Fewer Binary
        Variables*. IEEE Control Systems Letters, Under Review, 2022.


    .. warning::

        Drake must be compiled from source to support Gurobi and Mosek MICP solvers.
        See `<https://drake.mit.edu/from_source.html>`_ for more details.

        Drake's naive branch-and-bound solver does not require Gurobi or Mosek, and
        can be used with the ``bnb`` solver option, but this tends to be very slow. 

    :param spec:            An :class:`.STLFormula` describing the specification.
    :param sys:             A :class:`.LinearSystem` describing the system dynamics.
    :param x0:              A ``(n,1)`` numpy matrix describing the initial state.
    :param T:               A positive integer fixing the total number of timesteps :math:`T`.
    :param M:               (optional) A large positive scalar used to rewrite ``min`` and ``max`` as
                            mixed-integer constraints. Default is ``1000``.
    :param robustness_cost: (optional) Boolean flag for adding a linear cost to maximize
                            the robustness measure. Default is ``True``.
    :param solver:          (optional) String describing the solver to use. Must be one
                            of 'gurobi', 'mosek', or 'bnb'.
    :param presolve:        (optional) A boolean indicating whether to use gurobi's
                            presolve routines. Only affects the gurobi solver. Default is ``True``.
    :param verbose:         (optional) A boolean indicating whether to print detailed
                            solver info. Default is ``True``.
    """
    def __init__(self, spec, sys, x0, T, M=1000, robustness_cost=True, 
            solver='gurobi', presolve=True, verbose=True):
        super().__init__(spec, sys, x0, T, M, robustness_cost=robustness_cost, 
                solver=solver, presolve=presolve, verbose=verbose)

    def AddSubformulaConstraints(self, formula, z, t):
        """
        Given an STLFormula (formula) and a binary variable (z),
        add constraints to the optimization problem such that z
        takes value 1 only if the formula is satisfied (at time t).

        If the formula is a predicate, this constraint uses the "big-M"
        formulation

            A[x(t);u(t)] - b + (1-z)M >= 0,

        which enforces A[x;u] - b >= 0 if z=1, where (A,b) are the
        linear constraints associated with this predicate.

        If the formula is not a predicate, we recursively traverse the
        subformulas associated with this formula, adding new binary
        variables z_i for each subformula and constraining

            z <= z_i  for all i

        if the subformulas are combined with conjunction (i.e. all
        subformulas must hold), or otherwise constraining

            z <= sum(z_i)

        if the subformulas are combined with disjuction (at least one
        subformula must hold).
        """
        # We're at the bottom of the tree, so add the big-M constraints
        if isinstance(formula, LinearPredicate):
            # a.T*y - b + (1-z)*M >= rho
            y = self.y[:,t]
            self.mp.AddLinearConstraint(ge(
                formula.a.T@y - formula.b + (1-z)*self.M, self.rho
            ))

        # We haven't reached the bottom of the tree, so keep adding
        # boolean constraints recursively
        else:
            nz = len(formula.subformula_list)

            if formula.combination_type == "and":
                z_subs = self.mp.NewContinuousVariables(nz,1)
                self.mp.AddConstraint(le( z, z_subs ))

            else:  # combination_type == "or":
                lambda_, _ = AddLogarithmicSos1Constraint(self.mp, nz + 1)
                z_subs = lambda_[1:][np.newaxis].T
                self.mp.AddConstraint(eq( 1-z, lambda_[0] ))

            for i, subformula in enumerate(formula.subformula_list):
                t_sub = formula.timesteps[i]
                self.AddSubformulaConstraints(subformula, z_subs[i], t+t_sub)




from .drake_base import DrakeSTLSolver
from ...STL import LinearPredicate, NonlinearPredicate
import numpy as np

from pydrake.all import eq
from pydrake.solvers.all import IpoptSolver, SnoptSolver, SolverOptions, CommonSolverOption

import time

class DrakeSmoothSolver(DrakeSTLSolver):
    """
    Given an :class:`.STLFormula` :math:`\\varphi` and a :class:`.LinearSystem`,
    solve the optimization problem

    .. math::

        \max ~& \\rho^{\\varphi}(y_0,y_1,\dots,y_T)

        \\text{s.t. } & x_0 \\text{ fixed}

        & x_{t+1} = f(x_t, u_t)

        & y_{t} = g(x_t, u_t)

    where :math:`\\rho^{\\varphi}` is defined using a smooth approximation
    of min and max. We then use one of the general nonlinear solvers availible
    in Drake to find a locally optimal solution.

    .. Note::

        This class implements the algorithm described in

        Gilpin, Y, et al. *A Smooth Robustness Measure of Signal Temporal Logic for Symbolic Control*.
        IEEE Control Systems Letters, 2021.

    .. Note::

        This method is most effective when used in conjunction with the SNOPT
        sparse SQP solver. SNOPT is included with the binary version of Drake, but
        requires a license when built from source. 

    :param spec:    An :class:`.STLFormula` describing the specification.
    :param sys:     A :class:`.LinearSystem` describing the system dynamics.
    :param x0:      A ``(n,1)`` numpy matrix describing the initial state.
    :param T:       A positive integer fixing the total number of timesteps :math:`T`.
    :param k:       (optional). A smoothing parameter characterizing the tightness of
                    the smooth approximation. Larger values give a tighter approximation.
    :param verbose: (optional) A boolean indicating whether to print detailed
                    solver info. Default is ``True``.
    """

    def __init__(self, spec, sys, x0, T, k=2.0, verbose=True):
        DrakeSTLSolver.__init__(self, spec, sys, x0, T, verbose)
        self.k = k
        
        if self.verbose:
            print("Setting up optimization problem...")
            st = time.time()  # for computing setup time

        # Choose a solver
        snopt_solver = SnoptSolver()
        if snopt_solver.available():
            self.solver = snopt_solver
        else:
            print("\nWarning: this Drake installation does not include the sparse SQP solver SNOPT.")
            print("We'll try to use IPOPT instead, but this is slower and less reliable.")
            print("See https://drake.mit.edu/doxygen_cxx/classdrake_1_1solvers_1_1_snopt_solver.html for more details.\n")
            self.solver = IpoptSolver()

        # Add cost and constraints to the optimization problem
        self.AddDynamicsConstraints()
        self.AddSTLConstraints()
        self.AddRobustnessCost()

        if self.verbose:
            print(f"Setup complete in {time.time()-st} seconds.")

    def AddDynamicsConstraints(self):
        # Initial condition
        self.mp.AddConstraint(eq( self.x[:,0], self.x0 ))

        # Dynamics
        for t in range(self.T-1):
            self.mp.AddConstraint(eq(
                self.x[:,t+1], self.sys.f(self.x[:,t], self.u[:,t])
            ))
            self.mp.AddConstraint(eq(
                self.y[:,t], self.sys.g(self.x[:,t], self.u[:,t])
            ))
        self.mp.AddConstraint(eq(
            self.y[:,self.T-1], self.sys.g(self.x[:,self.T-1], self.u[:,self.T-1])
        ))

    def Solve(self):

        # Set solver options
        options = SolverOptions()
        if self.verbose:
            options.SetOption(CommonSolverOption.kPrintToConsole,1)
        self.mp.SetSolverOptions(options)

        # Local solvers tend to be sensitive to the initial guess
        np.random.seed(0)
        initial_guess = np.random.normal(size=self.mp.initial_guess().shape)

        st = time.time()
        res = self.solver.Solve(self.mp, initial_guess=initial_guess)
        solve_time = time.time() - st

        if res.is_success():
            if self.verbose:
                print("\nOptimal Solution Found!\n")
            x = res.GetSolution(self.x)
            u = res.GetSolution(self.u)

            # Report solve time and robustness
            y = self.sys.g(x, u)
            rho = self.spec.robustness(y,0)[0]
            if self.verbose:
                print("Solve time: ", solve_time)
                print("Optimal robustness: ", rho)

        else:
            if self.verbose:
                print("\nNo solution found.\n")
            x = None
            u = None
            rho = -np.inf

        return (x,u,rho,solve_time)

    def AddSTLConstraints(self):
        """
        Add the STL constraints

            (x,u) |= specification

        to the optimization problem, via the recursive introduction
        of binary variables for all subformulas in the specification.
        """
        # Recursively traverse the tree defined by the specification
        # to add constraints that define the STL robustness score
        self.AddSubformulaConstraints(self.spec, np.array([self.rho]), 0)

    def AddSubformulaConstraints(self, formula, rho, t):
        """
        Given an STLFormula (formula) and a continuous variable (rho),
        add constraints to the optimization problem such that rho is
        the robustness score for that formula.

        If the formula is a predicate (Ay-b>=0), this means that

            rho = A[x(t);u(t)] - b.

        If the formula is not a predicate, we recursively traverse the
        subformulas associated with this formula, adding new subformula
        robustness scores rho_i for each subformula and defining

            rho = min_i{ rho_i }

        if the subformulas are combined with conjunction and

            rho = max_i{ rho_i }

        if the subformulas are combined with disjuction.
        """
        # We're at the bottom of the tree, so add the predicate constraints
        if isinstance(formula, LinearPredicate):
            # rho = a'y - b
            y = self.y[:,t]
            self.mp.AddConstraint(eq( formula.a.T@y - formula.b, rho ))
        
        elif isinstance(formula, NonlinearPredicate):
            # rho = g(y)
            y = self.y[:,t]
            self.mp.AddConstraint(eq( formula.g(y), rho ))

        # We haven't reached the bottom of the tree, so keep adding
        # boolean constraints recursively
        else:
            rho_subs = []
            for i, subformula in enumerate(formula.subformula_list):
                rho_sub = self.mp.NewContinuousVariables(1)
                t_sub = formula.timesteps[i]   # the timestep at which this formula
                                               # should hold
                self.AddSubformulaConstraints(subformula, rho_sub, t+t_sub)
                rho_subs.append(rho_sub)

            if formula.combination_type == "and":
                # rho = min(rho_subs)
                self._add_min_constraint(rho, rho_subs)

            else:  # combination_type == "or":
                # rho = max(rho_subs)
                self._add_max_constraint(rho, rho_subs)

    def _add_max_constraint(self, a, b_lst):
        """
        Add constraints to the optimization problem such that

            a ~= max(b_1,b_2,...,b_N)

        where b_lst = [b_1,b_2,...,b_N] useing a smooth approximation of
        the max operator.
        """
        if len(b_lst) == 1:
            self.mp.AddConstraint(eq( a, b_lst[0] ))
        else:
            x = np.array(b_lst)
            exp = np.exp(self.k*x) + 1e-12   # avoid divide by zero error

            self.mp.AddConstraint(eq(
                a , np.sum(x*exp)/np.sum(exp)
            ))

    def _add_min_constraint(self, a, b_lst):
        """
        Add constraints to the optimization problem such that

            a ~= max(b_1,b_2,...,b_N)

        where b_lst = [b_1,b_2,...,b_N] useing a smooth approximation of
        the min operator.
        """
        if len(b_lst) == 1:
            self.mp.AddConstraint(eq( a, b_lst[0] ))
        else:
            x = np.hstack(b_lst)

            self.mp.AddConstraint(eq(
                a , -1./float(self.k) * np.log(np.sum(np.exp(-self.k*x)))
            ))




from abc import ABC, abstractmethod

class BenchmarkScenario(ABC):
    """
    An abstract base class defining a benchmark
    scenario for STL synthesis.
    """
    @abstractmethod
    def GetSpecification(self):
        """
        Get the STL specification associated with this scenario.

        :return spec: an ``STLFormula`` describing the specification.
        """
        pass
   
    @abstractmethod
    def GetSystem(self):
        """
        Get the system dynamics model associated with this
        scenario. 

        :return sys: a ``LinearSystem`` or ``NonlinearSystem`` 
                     specifying the system dynamics.
        """
        pass

    @abstractmethod
    def add_to_plot(self, ax):
        """
        Add a visualization of this specification
        to the given ``matplotlib`` axis.

        :param ax:  The ``matplotlib`` axis object to add the 
                    visualization to.
        """
        pass



from .base import BenchmarkScenario
from .common import (inside_rectangle_formula,
                     outside_rectangle_formula,
                     make_rectangle_patch)
from ..systems import DoubleIntegrator

class DoorPuzzle(BenchmarkScenario):
    r"""
    A mobile robot with double integrator dynamics must pick up
    several keys (by visiting certain regions :math:`\mathcal{K}_i`)
    before it cat pass through associated doors (:math:`\mathcal{D}_i`)
    and reach a goal (:math:`\mathcal{G}`). Along the way, it must
    avoid obstacles (:math:`\mathcal{O}_j`).

    .. math::

        \varphi = \bigwedge_{i=1}^{N} \left( \lnot \mathcal{D}_i U_{[0,T]} \mathcal{K}_i \right)  
        \land F_{[0,T]} \mathcal{G} 
        \land G_{[0,T]} (\bigwedge_{j=1}^5 \lnot \mathcal{O}_i)

    :param T:   The time horizon for this scenario
    :param N:   The number of key-door pairs. Must be between 1 and 4.
    """
    def __init__(self, T, N):
        assert N >= 1 and N <= 4, "number of pairs must be between 1 and 4"

        self.N = N
        self.T = T
   
        # Goal location
        self.goal_bounds = (14.1,14.9,4.1,5.9)

        # Obstacle locations
        self.obs1_bounds = (8,15.01,-0.01,4)
        self.obs2_bounds = (8,15.01,6,10.01)
        self.obs3_bounds = (3.5,5,-0.01,2.5)
        self.obs4_bounds = (-0.01,2.5,4,6)
        self.obs5_bounds = (3.5,5,7.5,10.01)

        # Door locations (set to overlap obstacles slightly)
        self.door1_bounds = (12.8,14,3.99,6.01)
        self.door2_bounds = (11.5,12.7,3.99,6.01)
        self.door3_bounds = (10.2,11.4,3.99,6.01)
        self.door4_bounds = (8.9,10.1,3.99,6.01)

        # Key locations
        self.key1_bounds = (1,2,1,2)
        self.key2_bounds = (1,2,8,9)
        self.key3_bounds = (6,7,8,9)
        self.key4_bounds = (6,7,1,2)

    def GetSpecification(self):
        # Goal Reaching
        at_goal = inside_rectangle_formula(self.goal_bounds, 0, 1, 6)

        # Obstacle Avoidance
        not_at_obs1 = outside_rectangle_formula(self.obs1_bounds, 0, 1, 6)
        not_at_obs2 = outside_rectangle_formula(self.obs2_bounds, 0, 1, 6)
        not_at_obs3 = outside_rectangle_formula(self.obs3_bounds, 0, 1, 6)
        not_at_obs4 = outside_rectangle_formula(self.obs4_bounds, 0, 1, 6)
        not_at_obs5 = outside_rectangle_formula(self.obs5_bounds, 0, 1, 6)
        obstacle_avoidance = not_at_obs1 & not_at_obs2 & not_at_obs3 & not_at_obs4 & not_at_obs5

        # Being outside a door region
        no_door1 = outside_rectangle_formula(self.door1_bounds, 0, 1, 6)
        no_door2 = outside_rectangle_formula(self.door2_bounds, 0, 1, 6)
        no_door3 = outside_rectangle_formula(self.door3_bounds, 0, 1, 6)
        no_door4 = outside_rectangle_formula(self.door4_bounds, 0, 1, 6)
       
        # Being in a key region
        key1 = inside_rectangle_formula(self.key1_bounds, 0, 1, 6)
        key2 = inside_rectangle_formula(self.key2_bounds, 0, 1, 6)
        key3 = inside_rectangle_formula(self.key3_bounds, 0, 1, 6)
        key4 = inside_rectangle_formula(self.key4_bounds, 0, 1, 6)
      
        # Not reaching a door until the correspoding key region has been
        # visited
        k1d1 = no_door1.until(key1, 0, self.T)
        k2d2 = no_door2.until(key2, 0, self.T)
        k3d3 = no_door3.until(key3, 0, self.T)
        k4d4 = no_door4.until(key4, 0, self.T)

        # Putting it all together
        if self.N == 1:
            key_constraints = k1d1
        elif self.N == 2:
            key_constraints = k1d1 & k2d2
        elif self.N == 3:
            key_constraints = k1d1 & k2d2 & k3d3
        elif elf.N == 4:
            key_constraints = k1d1 & k2d2 & k3d3 & k4d4

        # Put all of the constraints together in one specification
        specification = obstacle_avoidance.always(0, self.T) & \
                        key_constraints & \
                        at_goal.eventually(0, self.T)

        return specification

    def GetSystem(self):
        return DoubleIntegrator(2)

    def add_to_plot(self, ax):
        # Obstacles
        obs = [self.obs1_bounds, self.obs2_bounds, self.obs3_bounds, self.obs4_bounds, self.obs5_bounds]
        for obs_bounds in obs:
            obs_patch = make_rectangle_patch(*obs_bounds, color='k', alpha=0.5)
            ax.add_patch(obs_patch)
       
        # Doors
        door1 = make_rectangle_patch(*self.door1_bounds, color='red', alpha=0.5)
        door2 = make_rectangle_patch(*self.door2_bounds, color='red', alpha=0.5)
        door3 = make_rectangle_patch(*self.door3_bounds, color='red', alpha=0.5)
        door4 = make_rectangle_patch(*self.door4_bounds, color='red', alpha=0.5)

        # Keys
        key1 = make_rectangle_patch(*self.key1_bounds, color='blue', alpha=0.5)
        key2 = make_rectangle_patch(*self.key2_bounds, color='blue', alpha=0.5)
        key3 = make_rectangle_patch(*self.key3_bounds, color='blue', alpha=0.5)
        key4 = make_rectangle_patch(*self.key4_bounds, color='blue', alpha=0.5)

        # Goals
        goal = make_rectangle_patch(*self.goal_bounds, color='green', alpha=0.5)
        ax.add_patch(goal)

        # Only add the doors and keys that we actually used
        if self.N >= 1:
            ax.add_patch(door1)
            ax.add_patch(key1)

        if self.N >= 2:
            ax.add_patch(door2)
            ax.add_patch(key2)

        if self.N >= 3:
            ax.add_patch(door3)
            ax.add_patch(key3)

        if self.N >= 4:
            ax.add_patch(door4)
            ax.add_patch(key4)

        # set the field of view
        ax.set_xlim((0,15))
        ax.set_ylim((0,10))
        ax.set_aspect('equal')



import numpy as np
import time
from scipy.optimize import minimize
from ..base import STLSolver

class ScipyGradientSolver(STLSolver):
    """
    Given an :class:`.STLFormula` :math:`\\varphi` and a :class:`.NonlinearSystem`,
    solve the optimization problem

    .. math::

        \min &  - \\rho^{\\varphi}(y_0,y_1,\dots,y_T) + \sum_{t=0}^T x_t^TQx_t + u_t^TRu_t

        \\text{s.t. } & x_0 \\text{ fixed}

        & x_{t+1} = f(x_t, u_t)

        & y_{t} = g(x_t, u_t)

    using a shooting method and the
    `scipy.optimize <https://docs.scipy.org/doc/scipy/reference/tutorial/optimize.html>`_ solver.

    .. warning::

        This solver uses finite-differences to approximate the gradient of the (non-smooth) cost.
        As such, this method is likely to scale extremely poorly.

    :param spec:    An :class:`.STLFormula` describing the specification.
    :param sys:     A :class:`.NonlinearSystem` describing the system dynamics.
    :param x0:      A ``(n,1)`` numpy matrix describing the initial state.
    :param T:       A positive integer fixing the total number of timesteps :math:`T`.
    :param method:  (optional) String characterizing the optimization algorithm to use. See
                    `the scipy docs <https://docs.scipy.org/doc/scipy/reference/reference/generated/scipy.optimize.minimize.html>`_
                    for more details. Default is Sequential Least Squares (``"slsqp"``).
    :param verbose: (optional) A boolean indicating whether to print detailed
                    solver info. Default is ``True``.
    """
    def __init__(self, spec, sys, x0, T, method="slsqp", verbose=True):
        super().__init__(spec, sys, x0, T, verbose)
        self.Q = np.zeros((sys.n,sys.n))
        self.R = np.zeros((sys.m,sys.m))
        self.method = method

    def AddControlBounds(self, u_min, u_max):
        raise NotImplementedError("This solver does not support control bounds!")

    def AddStateBounds(self, x_min, x_max):
        raise NotImplementedError("This solver does not support state bounds!")

    def AddDynamicsConstraints(self):
        raise NotImplementedError("Dynamics constraints are added automatically in cost function computation")

    def AddQuadraticCost(self, Q, R):
        assert Q.shape == (self.sys.n, self.sys.n), "Q must be an (n,n) numpy array"
        assert R.shape == (self.sys.m, self.sys.m), "R must be an (m,m) numpy array"
        self.Q = Q
        self.R = R

    def AddRobustnessCost(self):
        raise NotImplementedError("Robustness cost is added automatically in cost function computation")

    def AddRobustnessConstraint(self):
        raise NotImplementedError("This solver does not support robustness constraints")

    def AddSTLConstraints(self):
        raise NotImplementedError("STL constraints are added automatically in cost function computation")

    def Solve(self):
        # Set an initial guess
        np.random.seed(0)  # for reproducability
        u_guess = np.random.uniform(-0.2,0.2,(self.sys.m,self.T))

        # Run scipy's minimize
        start_time = time.time()
        res = minimize(self.cost, u_guess.flatten(),
                method=self.method)
        solve_time = time.time() - start_time

        if self.verbose:
            print(res.message)
            print("Solve Time: ", solve_time)

        if res.success:
            u = res.x.reshape((self.sys.m,self.T))
            x, y = self.forward_rollout(u)

            rho = self.spec.robustness(y, 0)[0]
            if self.verbose:
                print("Optimal robustness: ", rho)
        else:
            x = None
            u = None
            rho = -np.inf

        return (x,u,rho,solve_time)

    def forward_rollout(self, u):
        """
        Given a control trajectory u of size (m,T),
        perform a forward rollout to compute the associated
        state and output trajectories.
        """
        T = u.shape[1]
        x = np.full((self.sys.n,T),np.nan)
        y = np.full((self.sys.p,T),np.nan)

        x[:,0] = self.x0

        for t in range(T-1):
            x[:,t+1] = self.sys.f(x[:,t], u[:,t])
            y[:,t] = self.sys.g(x[:,t], u[:,t])

        y[:,T-1] = self.sys.g(x[:,t-1], u[:,t-1])

        return x, y

    def cost(self, u_flat):
        """
        Compute the cost (negative robustness) associated
        with the (flattened) control sequence u.
        """
        cost = 0
        # Reconstruct the (m,T) control trajectory from the flattened
        # input. We use a flattened input because scipy's minimize
        # (and indeed most optimization software) assumes the decision
        # variables are put in a vector.
        u = u_flat.reshape((self.sys.m, self.T))

        # Do a forward rollout to compute the state and output trajectories
        x, y = self.forward_rollout(u)

        # Add additional state and control costs
        for t in range(self.T):
            cost += x[:,t].T@self.Q@x[:,t] + u[:,t].T@self.R@u[:,t]

        # Add the (negative) robustness of this signal y with respect
        # to the specification to the cost
        cost += -self.spec.robustness(y,0)

        return cost



# Test whether various optional dependencies are installed
DRAKE_ENABLED = True
try:
    import pydrake
except ImportError:
    print("WARNING: pydrake import failed, Drake-based solvers disabled.")
    print("         Install drake (https://drake.mit.edu/installation.html)")
    print("         to use the Drake-based solvers.")
    DRAKE_ENABLED = False

SCIPY_ENABLED = True
try:
    import scipy
except ImportError:
    print("WARNING: scipy import failed, scipy-based solvers disabled.")
    print("         Install scipy (https://scipy.org/install/)")
    print("         to use the scipy-based solvers.")
    SCIPY_ENABLED = False

GUROBI_ENABLED = True
try:
    import gurobipy
except ImportError:
    print("WARNING: gurobipy import failed, Gurobi-based solvers disabled.")
    print("         Install Gurobi (https://www.gurobi.com/documentation/)")
    print("         to use the Gurobi-based solvers. Free academic licenses")
    print("         are available.")
    GUROBI_ENABLED = False

# And load the corresponding solvers accordingly
if SCIPY_ENABLED:
    from .scipy.gradient_solver import ScipyGradientSolver

if GUROBI_ENABLED:
    from .gurobi.gurobi_micp import GurobiMICPSolver

if DRAKE_ENABLED:
    from .drake.drake_micp import DrakeMICPSolver
    from .drake.drake_smooth import DrakeSmoothSolver
    from .drake.drake_sos1 import DrakeSos1Solver



from .base import BenchmarkScenario
from .common import inside_circle_formula, make_circle_patch
from ..systems import Unicycle

class NonlinearReachAvoid(BenchmarkScenario):
    r"""
    A 2D mobile robot with unicycle dynamics must
    avoid a circular obstacle (:math:`\mathcal{O}`) before reaching 
    a circular goal (:math:`\mathcal{G}`):

    .. math::

        \varphi = G_{[0,T]} \lnot \mathcal{O} \land F_{[0,T]} \mathcal{G}

    :param goal_center:      a tuple ``(px, py)`` defining the center of the
                             goal region
    :param goal_radius:      a scalar defining the goal radius
    :param obstacle_center:  a tuple ``(px, py)`` defining the center of the
                             obstacle region
    :param obstacle_radius:  a scalar defining the obstacle radius
    :param T:                the time horizon for this scenario.
    """
    def __init__(self, goal_center, goal_radius, obstacle_center, obstacle_radius, T):
        self.goal_center = goal_center
        self.goal_radius = goal_radius

        self.obstacle_center = obstacle_center
        self.obstacle_radius = obstacle_radius

        self.T = T

    def GetSpecification(self):
        # Goal Reaching
        at_goal = inside_circle_formula(self.goal_center, self.goal_radius, 0, 1, 3)

        # Obstacle Avoidance
        at_obstacle = inside_circle_formula(self.obstacle_center,
                self.obstacle_radius, 0, 1, 3)
        not_at_obstacle = at_obstacle.negation()

        # Put all of the constraints together in one specification
        spec = not_at_obstacle.always(0, self.T) & at_goal.eventually(0, self.T)

        return spec

    def GetSystem(self):
        sys = Unicycle(dt=0.1)
        return sys

    def add_to_plot(self, ax):
        # Make and add circular patches
        obstacle = make_circle_patch(self.obstacle_center,
                self.obstacle_radius, color='k', alpha=0.5)
        goal = make_circle_patch(self.goal_center, self.goal_radius, 
                color='green', alpha=0.5)

        ax.add_patch(obstacle)
        ax.add_patch(goal)

        # set the field of view
        ax.set_xlim((0,10))
        ax.set_ylim((0,10))
        ax.set_aspect('equal')



from .base import BenchmarkScenario
from .common import (inside_rectangle_formula,
                     outside_rectangle_formula,
                     make_rectangle_patch)
from ..systems import DoubleIntegrator

class NarrowPassage(BenchmarkScenario):
    r"""
    A 2D mobile robot with double integrator dynamics must navigate around
    several obstacles (:math:`\mathcal{O}_i`) before reaching one of two
    goals (:math:`\mathcal{G}_i`).

    .. math::

        \varphi = F_{[0,T]}(\mathcal{G}_1 \lor \mathcal{G}_2) \land
            G_{[0,T]} \left( \bigwedge_{i=1}^4 \lnot \mathcal{O}_i \right)

    :param T:   The time horizon of the specification.
    """
    def __init__(self, T):
        # Define obstacle and goal regions by their bounds,
        # (xmin, xmax, ymin, ymax)
        self.obstacles = [(2,5,4,6),
                          (5.5,9,3.8,5.7),
                          (4.6,8,0.5,3.5),
                          (2.2,4.4,6.4,11)]
        self.goals = [(7,8,8,9),
                      (9.5,10.5,1.5,2.5)]
        self.T = T

    def GetSpecification(self):
        # Goal Reaching
        goal_formulas = []
        for goal in self.goals:
            goal_formulas.append(inside_rectangle_formula(goal, 0, 1, 6))

        at_any_goal = goal_formulas[0]
        for i in range(1,len(goal_formulas)):
            at_any_goal = at_any_goal | goal_formulas[i]

        # Obstacle Avoidance
        obstacle_formulas = []
        for obs in self.obstacles:
            obstacle_formulas.append(outside_rectangle_formula(obs, 0, 1, 6))

        obstacle_avoidance = obstacle_formulas[0]
        for i in range(1, len(obstacle_formulas)):
            obstacle_avoidance = obstacle_avoidance & obstacle_formulas[i]

        # Put all of the constraints together in one specification
        specification = at_any_goal.eventually(0, self.T) & \
                        obstacle_avoidance.always(0, self.T)

        return specification

    def GetSystem(self):
        return DoubleIntegrator(2)

    def add_to_plot(self, ax):
        # Make and add rectangular patches
        for obstacle in self.obstacles:
            ax.add_patch(make_rectangle_patch(*obstacle, color='k', alpha=0.5))
        for goal in self.goals:
            ax.add_patch(make_rectangle_patch(*goal, color='green', alpha=0.5))

        # set the field of view
        ax.set_xlim((0,12))
        ax.set_ylim((0,12))
        ax.set_aspect('equal')



from abc import ABC, abstractmethod

class STLSolver(ABC):
    """
    A simple abstract base class defining a common solver interface
    for different optimization-based STL synthesis methods.

    This class considers variations on the trajectory synthesis problem

    .. math::

        \min & \sum_{t=0}^T x_t^TQx_t + u_t^TRu_t

        \\text{s.t. } & x_0 \\text{ fixed }

        & x_{t+1} = f(x_t, u_t)

        & y_t = g(x_t, u_t)

        & \\rho^{\\varphi}(y_0,y_1,\dots,y_T) \geq 0

    where :math:`Q \\succeq 0` and :math:`R \\succeq 0` are cost weights, :math:`f` and :math:`g`
    define the system dynamics, and :math:`\\rho` is the robustness measure associated with
    a given STL specification :math:`\\varphi`.

    Possible variations include using the robustness measure :math:`\\rho^\\varphi`
    as a cost, dropping the quadratic running cost, and removing the hard satisfaction
    constriant :math:`\\rho^{\\varphi}\geq 0`.

    :param spec:    An :class:`.STLFormula` describing the specification.
    :param sys:     An :class:`.NonlinearSystem` characterizing the system dynamics.
    :param x0:      A ``(n,1)`` numpy array representing the initial state :math:`x_0`.
    :param T:       A positive integer fixing the total number of timesteps :math:`T`.
    :param verbose: A boolean specifying whether to print detailed solver info.
    """
    def __init__(self, spec, sys, x0, T, verbose):
        # Store the relevant data
        self.sys = sys
        self.spec = spec
        self.x0 = x0
        self.T = T+1  # needed to be consistent with how we've defined STLFormula
        self.verbose = verbose

    @abstractmethod
    def AddDynamicsConstraints(self):
        """
        Add the dynamics constraints

        .. math::

            & x_0 \\text{ fixed }

            & x_{t+1} = f(x_t, u_t)

            & y_t = g(x_t, u_t)

        to the optimization problem.
        """
        pass

    @abstractmethod
    def AddSTLConstraints(self):
        """
        Add constraints to the optimization problem to define
        the robustness measure

        .. math::

            \\rho^{\\varphi}(y_0,y_1,\dots,y_T).
        """
        pass

    @abstractmethod
    def AddRobustnessCost(self):
        """
        Add the robustness measure as a (linear) cost
        to the optimization problem:

        .. math::

            \min -\\rho^{\\varphi}(y_0,y_1,\dots,y_T).

        """
        pass

    @abstractmethod
    def AddRobustnessConstraint(self, rho_min=0.0):
        """
        Add a constraint on the robustness measure to the
        optimization problem:

        .. math::

            \\rho^{\\varphi}(y_0,y_1,\dots,y_T) \geq \\rho_{min}

        :param rho_min:     (optional) Minimum robustness measure :math:`\\rho_{min}`.
                            Defaults to 0, which enforces STL satisfaction.
        """
        pass

    @abstractmethod
    def AddControlBounds(self, u_min, u_max):
        """
        Add upper and lower bounds on the control inputs :math:`u_t`
        to the optimization problem:

        .. math::

            u_{min} \leq u_t \leq u_{max} \quad \\forall t

        :param u_min:   A ``(m,)`` numpy array specifying the minimum control input
        :param u_max:   A ``(m,)`` numpy array specifying the maximum control input
        """
        pass

    @abstractmethod
    def AddStateBounds(self, x_min, x_max):
        """
        Add upper and lower bounds on the state variables :math:`x_t`
        to the optimization problem:

        .. math::

            x_{min} \leq x_t \leq x_{max} \quad \\forall t

        :param x_min:   A ``(n,)`` numpy array specifying :math:`x_{min}`
        :param x_max:   A ``(n,)`` numpy array specifying :math:`x_{max}`
        """
        pass

    @abstractmethod
    def AddQuadraticCost(self, Q, R):
        """
        Add a quadratic running cost to the optimization problem:

        .. math::

            \min \sum_{t=0}^T x_t^TQx_t + u_t^TRu_t

        :param Q:   A ``(n,n)`` numpy array representing the state penalty matrix :math:`Q`
        :param R:   A ``(m,m)`` numpy array representing the control penalty matrix :math:`R`
        """
        pass

    @abstractmethod
    def Solve(self):
        """
        Solve the STL syntheis optimization problem and return an optimal trajectory.

        :return x:          A ``(n,T)`` numpy array containing the optimal state :math:`x_t`
                            for each timestep.
        :return u:          A ``(m,T)`` numpy array containing the optimal control :math:`x_t`
                            for each timestep.
        :return rho:        A scalar indicating the optimal robustness value.
        :return solve_time: The time it took the solver to find a solution, in seconds.

        .. note::

            ``x`` and ``u`` are returned as ``None`` if the optimization problem is
            infeasible or the solver is unable to find a solution.
        """
        pass




from setuptools import setup, find_packages

long_description = """
A python library for control from Signal Temporal Logic (STL) specifications.

This software is designed with the following goals in mind:

- Provide a simple python interface for dealing with STL formulas
- Provide high-quality implementations of several state-of-the-art synthesis 
  algorithms, including Mixed-Integer Convex Programming (MICP) and 
  gradient-based optimization.
- Make it easy to design and evaluate new synthesis algorithms.
- Provide a variety of benchmark scenarios that can be used to test new algorithms.
"""

setup(name='stlpy',
      version='0.2.1',
      description='A Python library for control from Signal Temporal Logic (STL) specifications',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://stlpy.readthedocs.io/en/latest/index.html',
      project_urls ={
          "Source Code": "https://github.com/vincekurtz/stlpy",
          "Documentation": "https://stlpy.readthedocs.io/en/latest/index.html"},
      author='Vince Kurtz',
      author_email='vjkurtz@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=['numpy', 'matplotlib', 'treelib'],
      python_requires='>=3.8',
      zip_safe=False)



import numpy as np
from abc import ABC, abstractmethod
from treelib import Tree

class STLFormula(ABC):
    """
    An abstract class which encompasses represents all kinds of STL formulas :math:`\\varphi`, including
    predicates (the simplest possible formulas) and standard formulas (made up of logical operations over
    predicates and other formulas).
    """
    @abstractmethod
    def robustness(self, y, t):
        """
        Compute the robustness measure :math:`\\rho^\\varphi(y,t)` of this formula for the
        given signal :math:`y = y_0,y_1,\\dots,y_T`, evaluated at timestep :math:`t`.

        :param y:    A ``(d,T)`` numpy array representing the signal
                     to evaluate, where ``d`` is the dimension of
                     the signal and ``T`` is the number of timesteps
        :param t:    The timestep :math:`t` to evaluate the signal at. This
                     is typically 0 for the full formula.

        :return:    The robustness measure :math:`\\rho^\\varphi(y,t)` which is positive only
                    if the signal satisfies the specification.
        """
        pass

    @abstractmethod
    def is_predicate(self):
        """
        Indicate whether this formula is a predicate.

        :return:    A boolean which is ``True`` only if this is a predicate.
        """
        pass

    @abstractmethod
    def is_state_formula(self):
        """
        Indicate whether this formula is a state formula, e.g.,
        a predicate or the result of boolean operations over
        predicates.

        :return:    A boolean which is ``True`` only if this is a state formula.
        """
        pass

    @abstractmethod
    def is_disjunctive_state_formula(self):
        """
        Indicate whether this formula is a state formula defined by
        only disjunctions (or) over predicates.

        :return:     A boolean which is ``True`` only if this is a disjunctive state formula.
        """
        pass

    @abstractmethod
    def is_conjunctive_state_formula(self):
        """
        Indicate whether this formula is a state formula defined by
        only conjunctions (and) over predicates.

        :return:     A boolean which is ``True`` only if this is a conjunctive state formula.
        """
        pass

    @abstractmethod
    def get_all_inequalities(self):
        """
        Return all inequalities associated with this formula stacked into vector form

        .. math::

            Ay \le b

        where each row of :math:`A` and :math:`b` correspond to a predicate in this formula.

        .. note::

            This method is really only useful for conjunctive state formulas.

        :return A:  An (n,m) numpy array representing :math:`A`
        :return b:  An (n,) numpy array representing :math:`b`
        """
        pass

    @abstractmethod
    def negation(self):
        """
        Return a new :class:`.STLFormula` :math:`\\varphi_{new}` which represents
        the negation of this formula:

        .. math::

            \\varphi_{new} = \lnot \\varphi

        :return: An :class:`.STLFormula` representing :math:`\\varphi_{new}`

        .. note::

            For now, only formulas in positive normal form are supported. That means that negation
            (:math:`\lnot`) can only be applied to predicates (:math:`\\pi`).

        """
        pass

    def conjunction(self, other):
        """
        Return a new :class:`.STLTree` :math:`\\varphi_{new}` which represents the conjunction
        (and) of this formula (:math:`\\varphi`) and another one (:math:`\\varphi_{other}`):

        .. math::

            \\varphi_{new} = \\varphi \land \\varphi_{other}

        :param other:   The :class:`.STLFormula` :math:`\\varphi_{other}`

        :return: An :class:`.STLTree` representing :math:`\\varphi_{new}`

        .. note::

            Conjuction can also be represented with the ``&`` operator, i.e.,
            ::

                c = a & b

            is equivalent to
            ::

                c = a.conjuction(b)

        """
        return STLTree([self,other],"and",[0,0])

    def __and__(self, other):
        """
        Syntatic sugar so we can write `one_and_two = one & two`
        """
        return self.conjunction(other)

    def disjunction(self, other):
        """
        Return a new :class:`.STLTree` :math:`\\varphi_{new}` which represents the disjunction
        (or) of this formula (:math:`\\varphi`) and another one (:math:`\\varphi_{other}`):

        .. math::

            \\varphi_{new} = \\varphi \lor \\varphi_{other}

        :param other:   The :class:`.STLFormula` :math:`\\varphi_{other}`

        :return: An :class:`.STLTree` representing :math:`\\varphi_{new}`

        .. note::

            Disjunction can also be represented with the ``|`` operator, i.e.,
            ::

                c = a | b

            is equivalent to
            ::

                c = a.disjunction(b)

        """
        return STLTree([self,other],"or",[0,0])

    def __or__(self, other):
        """
        Syntatic sugar so we can write `one_or_two = one | two`
        """
        return self.disjunction(other)

    def always(self, t1, t2):
        """
        Return a new :class:`.STLTree` :math:`\\varphi_{new}` which ensures that this
        formula (:math:`\\varphi`) holds for all of the timesteps between
        :math:`t_1` and :math:`t_2`:

        .. math::

            \\varphi_{new} = G_{[t_1,t_2]}(\\varphi)


        :param t1:  An integer representing the delay :math:`t_1`
        :param t2:  An integer representing the deadline :math:`t_2`

        :return: An :class:`.STLTree` representing :math:`\\varphi_{new}`
        """
        time_interval = [t for t in range(t1,t2+1)]
        subformula_list = [self for t in time_interval]
        formula = STLTree(subformula_list, "and", time_interval)
        if self.name is not None:
            formula.name = "always [%s,%s] %s" % (t1,t2,self.name)
        return formula

    def eventually(self, t1, t2):
        """
        Return a new :class:`.STLTree` :math:`\\varphi_{new}` which ensures that this
        formula (:math:`\\varphi`) holds for at least one timestep between
        :math:`t_1` and :math:`t_2`:

        .. math::

            \\varphi_{new} = F_{[t_1,t_2]}(\\varphi)


        :param t1:  An integer representing the delay :math:`t_1`
        :param t2:  An integer representing the deadline :math:`t_2`

        :return: An :class:`.STLTree` representing :math:`\\varphi_{new}`
        """
        time_interval = [t for t in range(t1,t2+1)]
        subformula_list = [self for t in time_interval]
        formula = STLTree(subformula_list, "or", time_interval)
        if self.name is not None:
            formula.name = "eventually [%s,%s] %s" % (t1,t2,self.name)
        return formula

    def until(self, other, t1, t2):
        """
        Return a new :class:`.STLTree` :math:`\\varphi_{new}` which ensures that the
        given formula :math:`\\varphi_{other}` holds for at least one timestep between
        :math:`t_1` and :math:`t_2`, and that this formula (:math:`\\varphi`) holds
        at all timesteps until then:

        .. math::

            \\varphi_{new} = \\varphi U_{[t_1,t_2]}(\\varphi_{other})

        :param other:   A :class:`.STLFormula` representing :math:`\\varphi_{other`
        :param t1:  An integer representing the delay :math:`t_1`
        :param t2:  An integer representing the deadline :math:`t_2`

        :return: An :class:`.STLTree` representing :math:`\\varphi_{new}`
        """
        # For every candidate swiching time (t_prime), construct a subformula
        # representing 'self' holding until t_prime, at which point 'other' holds.
        self_until_tprime = []

        for t_prime in range(t1, t2+1):
            time_interval = [t for t in range(t1, t_prime+1)]
            subformula_list = [self for t in range(t1, t_prime)]
            subformula_list.append(other)
            self_until_tprime.append(STLTree(subformula_list, "and", time_interval))

        # Then we take the disjunction over each of these formulas
        return STLTree(self_until_tprime, "or", [0 for i in range(len(self_until_tprime))])

    def get_all_conjunctive_state_formulas(self):
        """
        Return a list of all of the (unique) conjunctive state
        formulas that make up this specification.

        :return:    A list of STLFormula objects
        """
        CSFs = []
        if self.is_conjunctive_state_formula():
            CSFs.append(self)
        else:
            for subformula in self.subformula_list:
                new_csf_list = subformula.get_all_conjunctive_state_formulas()

                # TODO: Deal with special case where there is only one formula in the
                # conjunction list: this comes up when the UNTIL operator is used

                # Only add those formulas that we don't already have to the list
                for csf in new_csf_list:
                    if not csf in CSFs:
                        CSFs.append(csf)

        return CSFs

class STLTree(STLFormula):
    """
    Describes an STL formula :math:`\\varphi` which is made up of
    operations over :class:`.STLFormula` objects. This defines a tree structure,
    so that, for example, the specification

    .. math::

        \\varphi = G_{[0,3]} \\pi_1 \land F_{[0,3]} \\pi_2

    is represented by the tree

    .. graphviz::

        digraph tree {
            root [label="phi"];
            G [label="G"];
            F [label="F"];
            n1 [label="pi_1"];
            n2 [label="pi_1"];
            n3 [label="pi_1"];
            n4 [label="pi_2"];
            n5 [label="pi_2"];
            n6 [label="pi_2"];

            root -> G;
            root -> F;
            G -> n1;
            G -> n2;
            G -> n3;
            F -> n4;
            F -> n5;
            F -> n6;
        }

    where each node is an :class:`.STLFormula` and the leaf nodes are :class:`.LinearPredicate` objects.


    Each :class:`.STLTree` is defined by a list of :class:`.STLFormula` objects
    (the child nodes in the tree) which are combined together using either conjunction or
    disjunction.

    :param subformula_list:     A list of :class:`.STLFormula` objects (formulas or
                                predicates) that we'll use to construct this formula.
    :param combination_type:    A string representing the type of operation we'll use
                                to combine the child nodes. Must be either ``"and"`` or ``"or"``.
    :param timesteps:           A list of timesteps that the subformulas must hold at.
                                This is needed to define the temporal operators.
    """
    def __init__(self, subformula_list, combination_type, timesteps, name=None):
        # Record the dimension of the signal this formula is defined over
        self.d = subformula_list[0].d

        # Run some type check on the inputs
        assert (combination_type == "and") or (combination_type == "or"), "Invalid combination type"
        assert isinstance(subformula_list, list), "subformula_list must be a list of STLTree or LinearPredicate objects"
        assert isinstance(timesteps, list), "timesteps must be a list of integers"
        assert len(timesteps) == len(subformula_list), "a timestep must be provided for each subformula"
        for formula in subformula_list:
            assert isinstance(formula, STLFormula), "subformula_list must be a list of STLTree or LinearPredicate objects"
            assert formula.d == self.d, "all subformulas must be defined over same dimension of signal"
        for t in timesteps:
            assert isinstance(t, int), "each timestep must be an integer"

        # Simply save the input arguments. We will parse these recursively later on to
        # determine, for example, the formula robustness.
        self.subformula_list = subformula_list
        self.combination_type = combination_type
        self.timesteps = timesteps

        # Save the given name for pretty printing
        self.name=name

    def negation(self):
        raise NotImplementedError("Only formulas in positive normal form are supported at this time")

    def robustness(self, y, t):
        if self.combination_type == "and":
            return min( [formula.robustness(y,t+self.timesteps[i]) for i, formula in enumerate(self.subformula_list)] )
        else: # combination_type == "or"
            return max( [formula.robustness(y,t+self.timesteps[i]) for i, formula in enumerate(self.subformula_list)] )

    def is_predicate(self):
        return False

    def is_state_formula(self):
        boolean_operation = all([self.timesteps[i] == self.timesteps[0] for i in range(len(self.timesteps))])
        children_are_state_formulas = all([subformula.is_state_formula() for subformula in self.subformula_list])

        return boolean_operation and children_are_state_formulas

    def is_disjunctive_state_formula(self):
        boolean_operation = all([self.timesteps[i] == self.timesteps[0] for i in range(len(self.timesteps))])
        children_match = all([s.is_disjunctive_state_formula() for s in self.subformula_list])

        return boolean_operation and children_match and self.combination_type == "or"

    def is_conjunctive_state_formula(self):
        boolean_operation = all([self.timesteps[i] == self.timesteps[0] for i in range(len(self.timesteps))])
        children_match = all([s.is_conjunctive_state_formula() for s in self.subformula_list])

        return boolean_operation and children_match and self.combination_type == "and"

    def simplify(self):
        """
        Modify this formula to reduce the depth of the formula tree while preserving
        logical equivalence.

        A shallower formula tree can result in a more efficient binary encoding in some
        cases.
        """
        mod = True
        while mod:
            # Just keep trying to flatten until we don't get any improvement
            mod = self.flatten(self)

    def flatten(self, formula):
        """
        Reduce the depth of the given :class:`STLFormula` by combining adjacent
        layers with the same logical operation. This preserves the meaning of the
        formula, since, for example,

        ..math::

            (a \land b) \land (c \land d) = a \land b \land c \land d

        :param formula: The formula to modify

        :return made_modification: boolean flag indicating whether the formula was changed.l

        """
        made_modification = False

        for subformula in formula.subformula_list:
            if subformula.is_predicate():
                pass
            else:
                if formula.combination_type == subformula.combination_type:
                    # Remove the subformula
                    i = formula.subformula_list.index(subformula)
                    formula.subformula_list.pop(i)
                    st = formula.timesteps.pop(i)

                    # Add all the subformula's subformulas instead
                    formula.subformula_list += subformula.subformula_list
                    formula.timesteps += [t+st for t in subformula.timesteps]
                    made_modification = True

                made_modification = self.flatten(subformula) or made_modification

        return made_modification

    def get_all_inequalities(self):
        As = []
        bs = []
        for subformula in self.subformula_list:
            A, b = subformula.get_all_inequalities()
            As.append(A)
            bs.append(b)
        A = np.vstack(As)
        b = np.hstack(bs)

        return A, b

    def __str__(self):
        """
        Return a string representing this formula. This string displays
        the tree structure of the formula, where each node represents either
        a conjuction or disjuction of subformulas, and leaves are state formulas.
        """
        tree = Tree()
        root = tree.create_node(self.combination_type)

        for subformula in self.subformula_list:
            self._add_subformula_to_tree(tree, root, subformula)

        return tree.__str__()

    def _add_subformula_to_tree(self, tree, root, formula):
        """
        Helper function for recursively parsing subformulas to create
        a Tree object for visualizing this formula.
        """
        if formula.is_predicate():
            tree.create_node(formula.__str__(), parent=root)
        else:
            new_node = tree.create_node(formula.combination_type, parent=root)
            for subformula in formula.subformula_list:
                self._add_subformula_to_tree(tree, new_node, subformula)



import numpy as np

from .base import BenchmarkScenario
from .common import inside_rectangle_formula, make_rectangle_patch
from ..systems import DoubleIntegrator

class SteppingStones(BenchmarkScenario):
    r"""
    A 2D mobile robot with double integrator dynamics must
    navigate to a goal (:math:`\mathcal{G}`) while only stepping 
    on certain predefined spaces (:math:`\mathcal{S}_i`):

    .. math::

        \varphi = G_{[0,T]} \left( \bigvee_{i}^{N_s} \mathcal{S}_i \right) \land
                  F_{[0,T]} \mathcal{G}

    :param num_stones:  The number of stepping stones :math:`N_s`
    :param T:           The specification time horizon
    :param seed:        (optional) The random seed for stone placement.
                        Default is None.
    """
    def __init__(self, num_stones, T, seed=None):
        self.T = T

        # Set the seed for the random number generator (for reproducability)
        np.random.seed(seed)

        # Create the (randomly generated) set of stepping stones
        self.stones = []
        for i in range(num_stones):
            x = np.random.uniform(0,9)  # keep within workspace
            y = np.random.uniform(0,9)
            self.stones.append((x,x+1,y,y+1))

        # Specify the target/goal
        self.target = self.stones[-1]

    def GetSpecification(self):

        # Specify that we must be on any one of the stones
        stone_formulas = []
        for stone in self.stones:
            stone_formulas.append(inside_rectangle_formula(stone, 0, 1, 6))

        on_any_stone = stone_formulas[0]
        for i in range(1, len(stone_formulas)):
            on_any_stone = on_any_stone | stone_formulas[i]

        # Specify that we much reach the target
        reach_target = inside_rectangle_formula(self.target, 0, 1, 6)

        # Put all of the constraints together in one specification
        specification = on_any_stone.always(0, self.T) & \
                        reach_target.eventually(0, self.T)

        return specification

    def GetSystem(self):
        return DoubleIntegrator(2)

    def add_to_plot(self, ax):
        n_stones = len(self.stones)

        # Add rectangles for the stones
        for i in range(n_stones):
            if i < n_stones-1:  # ordinary stepping stones are orange
                ax.add_patch(make_rectangle_patch(*self.stones[i], color='orange', alpha=0.5, zorder=-1))
            else:  # the target is green
                ax.add_patch(make_rectangle_patch(*self.stones[i], color='g', alpha=0.5, zorder=-1))

        # set the field of view
        ax.set_xlim((0,10))
        ax.set_ylim((0,10))
        ax.set_aspect('equal')



from .reach_avoid import ReachAvoid
from .either_or import EitherOr
from .narrow_passage import NarrowPassage
from .random_multitarget import RandomMultitarget
from .door_puzzle import DoorPuzzle
from .stepping_stones import SteppingStones
from .nonlinear_reach_avoid import NonlinearReachAvoid



from .drake_base import DrakeSTLSolver
from ...STL import LinearPredicate, NonlinearPredicate
import numpy as np
import time
from pydrake.all import (GurobiSolver, MosekSolver, ClpSolver,
                         SolverOptions, CommonSolverOption,
                         eq, le, ge)
from pydrake.solvers.branch_and_bound import MixedIntegerBranchAndBound

class DrakeMICPSolver(DrakeSTLSolver):
    """
    Given an :class:`.STLFormula` :math:`\\varphi` and a :class:`.LinearSystem`,
    solve the optimization problem

    .. math::

        \min & -\\rho^{\\varphi}(y_0,y_1,\dots,y_T) + \sum_{t=0}^T x_t^TQx_t + u_t^TRu_t

        \\text{s.t. } & x_0 \\text{ fixed}

        & x_{t+1} = A x_t + B u_t

        & y_{t} = C x_t + D u_t

        & \\rho^{\\varphi}(y_0,y_1,\dots,y_T) \geq 0

    using mixed-integer convex programming. This gives a globally optimal
    solution, but may be computationally expensive for long and complex specifications.

    .. note::

        This class implements the algorithm described in

        Belta C, et al.
        *Formal methods for control synthesis: an optimization perspective*.
        Anual Review of Control, Robotics, and Autonomous Systems, 2019.

    .. warning::

        Drake must be compiled from source to support Gurobi and Mosek MICP solvers.
        See `<https://drake.mit.edu/from_source.html>`_ for more details. 

        Drake's naive branch-and-bound solver does not require Gurobi or Mosek, and
        can be used with the ``bnb`` solver option, but this tends to be very slow. 

    :param spec:            An :class:`.STLFormula` describing the specification.
    :param sys:             A :class:`.LinearSystem` describing the system dynamics.
    :param x0:              A ``(n,1)`` numpy matrix describing the initial state.
    :param T:               A positive integer fixing the total number of timesteps :math:`T`.
    :param M:               (optional) A large positive scalar used to rewrite ``min`` and ``max`` as
                            mixed-integer constraints. Default is ``1000``.
    :param robustness_cost: (optional) Boolean flag for adding a linear cost to maximize
                            the robustness measure. Default is ``True``.
    :param solver:          (optional) String describing the solver to use. Must be one
                            of 'gurobi', 'mosek', or 'bnb'.
    :param presolve:        (optional) A boolean indicating whether to use gurobi's
                            presolve routines. Only affects the gurobi solver. Default is ``True``.
    :param verbose:         (optional) A boolean indicating whether to print detailed
                            solver info. Default is ``True``.
    """
    def __init__(self, spec, sys, x0, T, M=1000, robustness_cost=True, 
            solver='gurobi', presolve=True, verbose=True):
        assert M > 0, "M should be a (large) positive scalar"
        super().__init__(spec, sys, x0, T, verbose)

        self.M = M
        self.presolve = presolve

        # Choose which solver to use
        if solver == 'gurobi':
            self.solver = GurobiSolver()
        elif solver == 'mosek':
            self.solver = MosekSolver()
        else:
            print("Using Naive Branch-and-Bound solver")
            self.solver = "bnb"

        # Add cost and constraints to the optimization problem
        self.AddDynamicsConstraints()
        self.AddSTLConstraints()
        self.AddRobustnessConstraint()
        if robustness_cost:
            self.AddRobustnessCost()

    def Solve(self):

        # Set solver options
        options = SolverOptions()
        if self.verbose:
            options.SetOption(CommonSolverOption.kPrintToConsole,1)
        if not self.presolve:
            options.SetOption(GurobiSolver.id(), "Presolve", 0)
        self.mp.SetSolverOptions(options)
        
        # Setup for naive branch and bound
        if self.solver == "bnb":
            bnb_solver = MixedIntegerBranchAndBound(self.mp, ClpSolver.id())
            st = time.time()
            status = bnb_solver.Solve()
            solve_time = time.time() - st
            success = True
            res = bnb_solver

        # Setup for Gurobi/Mosek
        else:
            res = self.solver.Solve(self.mp)
            success = res.is_success()
            solve_time = res.get_solver_details().optimizer_time

        if self.verbose:
            print("")
            print("Solve time: ", solve_time)

        if success:
            x = res.GetSolution(self.x)
            u = res.GetSolution(self.u)

            y = self.sys.C@x + self.sys.D@u
            rho = self.spec.robustness(y,0)[0]
            if self.verbose:
                print("Optimal robustness: ", rho)
        else:
            if self.verbose:
                print("No solution found")
            x = None
            u = None
            rho = -np.inf

        return (x,u, rho, solve_time)

    def AddDynamicsConstraints(self):
        """
        Add the constraints

            x_{t+1} = A@x_t + B@u_t
            x_0 = x0

        to the optimization problem.
        """
        # Initial condition
        self.mp.AddConstraint(eq( self.x[:,0], self.x0 ))

        # Dynamics
        for t in range(self.T-1):
            self.mp.AddConstraint(eq(
                self.x[:,t+1], self.sys.A@self.x[:,t] + self.sys.B@self.u[:,t]
            ))
            self.mp.AddConstraint(eq(
                self.y[:,t], self.sys.C@self.x[:,t] + self.sys.D@self.u[:,t]
            ))
        self.mp.AddConstraint(eq(
            self.y[:,self.T-1], self.sys.C@self.x[:,self.T-1] + self.sys.D@self.u[:,self.T-1]
        ))

    def AddSTLConstraints(self):
        """
        Add the STL constraints

            (x,u) |= specification

        to the optimization problem, via the recursive introduction
        of binary variables for all subformulas in the specification.
        """
        # Add a binary variable which takes a value of 1 only
        # if the overall specification is satisfied.
        z_spec = self.mp.NewContinuousVariables(1)
        self.mp.AddConstraint(eq( z_spec, 1 ))

        # Recursively traverse the tree defined by the specification
        # subformulas and add similar binary constraints.
        self.AddSubformulaConstraints(self.spec, z_spec, 0)

    def AddSubformulaConstraints(self, formula, z, t):
        """
        Given an STLFormula (formula) and a binary variable (z),
        add constraints to the optimization problem such that z
        takes value 1 only if the formula is satisfied (at time t).

        If the formula is a predicate, this constraint uses the "big-M"
        formulation

            A[x(t);u(t)] - b + (1-z)M >= 0,

        which enforces A[x;u] - b >= 0 if z=1, where (A,b) are the
        linear constraints associated with this predicate.

        If the formula is not a predicate, we recursively traverse the
        subformulas associated with this formula, adding new binary
        variables z_i for each subformula and constraining

            z <= z_i  for all i

        if the subformulas are combined with conjunction (i.e. all
        subformulas must hold), or otherwise constraining

            z <= sum(z_i)

        if the subformulas are combined with disjuction (at least one
        subformula must hold).
        """
        # We're at the bottom of the tree, so add the big-M constraints
        if isinstance(formula, LinearPredicate):
            # a.T*y - b + (1-z)*M >= rho
            y = self.y[:,t]
            self.mp.AddLinearConstraint(ge(
                formula.a.T@y - formula.b + (1-z)*self.M, self.rho
            ))

            # By default, z is defined as a continuous variable. This
            # is a hack to make it binary. The redundant continuous
            # variable will be removed in presolve.
            tmp = self.mp.NewBinaryVariables(1)
            self.mp.AddConstraint(eq(tmp, z))

        elif isinstance(formula, NonlinearPredicate):
            raise TypeError("Mixed integer programming does not support nonlinear predicates")

        # We haven't reached the bottom of the tree, so keep adding
        # boolean constraints recursively
        else:
            z_subs = self.mp.NewContinuousVariables(len(formula.subformula_list),1)
            self.mp.AddConstraint(ge(z_subs, 0))

            if formula.combination_type == "and":
                self.mp.AddConstraint(le( z, z_subs ))
            else:  # combination_type == "or":
                self.mp.AddConstraint(le( z, sum(z_subs) ))

            for i, subformula in enumerate(formula.subformula_list):
                t_sub = formula.timesteps[i]
                self.AddSubformulaConstraints(subformula, z_subs[i], t+t_sub)
